/**
 * Copyright 2010 Andy Turner, The University of Leeds, UK
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package uk.ac.leeds.ccg.andyt.generic.math;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.util.TreeMap;

public class Generic_BigInteger extends Generic_Number {

    public static final BigInteger Two = BigInteger.valueOf(2);
    public static final BigInteger Three = BigInteger.valueOf(3);
    public static final BigInteger Hundred = BigInteger.valueOf(100);
    public static final BigInteger Integer_MAX_VALUE = BigInteger.valueOf(Integer.MAX_VALUE);
    public static final BigInteger Long_MAX_VALUE = BigInteger.valueOf(Long.MAX_VALUE);
    public TreeMap<Integer, BigInteger> _Factorial_TreeMap;

    /** Creates a new instance of Generic_BigDecimal */
    public Generic_BigInteger() {
        //super();
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        new Generic_BigInteger().test();
    }

    private void test() {
        test_power();
        //test_factorial();
    }

    public void test_power() {
        System.out.println("test_power()");
        BigInteger x;
        BigDecimal y;
        BigDecimal power;
        int decimalPlaces;
        RoundingMode a_RoundingMode = RoundingMode.HALF_UP;
        long time;
        decimalPlaces = 10;

        int n;
        x = new BigInteger("5");
        n = 13;
        time = System.currentTimeMillis();
        power = power(
                x,
                n,
                decimalPlaces,
                a_RoundingMode);
        time = timeReport(time);
        System.out.println("power(" + x + "," + n + ") " + power.toPlainString());

        x = new BigInteger("2");
        n = 256;
        for (int i = 1; i < 20; i++) {
            n *= 16;
            time = System.currentTimeMillis();
            power = power(
                    x,
                    n,
                    decimalPlaces,
                    a_RoundingMode);
            time = timeReport(time);
            System.out.println("power(" + x + "," + n + ") " + power.toPlainString());
        }

    }

    public void test_factorial() {
        int x;
        x = 10;
        BigInteger factorial;
        factorial = factorial(x);
        System.out.println("factorial(" + x + ") " + factorial.toString());
        x = 10000;
        factorial = factorial(x);
        System.out.println("factorial(" + x + ") " + factorial.toString());
    }

    /**
     * Initialises _Factorial_TreeMap
     */
    public void init_Factorial_TreeMap() {
        _Factorial_TreeMap = new TreeMap<Integer, BigInteger>();
        _Factorial_TreeMap.put(0, BigInteger.ONE);
        _Factorial_TreeMap.put(1, BigInteger.ONE);
    }

    /**
     * @param x
     * @param y
     * @return the larger (closer to positive infinity) of either x or y
     */
    public static BigInteger max(
            BigInteger x,
            BigInteger y) {
        if (x.compareTo(y) == -1) {
            return new BigInteger(y.toString());
        } else {
            return new BigInteger(x.toString());
        }
    }

    /**
     * @param x
     * @param y
     * @return the larger (closer to negative infinity) of either x or y
     */
    public static BigInteger min(
            BigInteger x,
            BigInteger y) {
        if (x.compareTo(y) == -1) {
            return new BigInteger(x.toString());
        } else {
            return new BigInteger(y.toString());
        }
    }

    /**
     * Adds values to _Factorial_TreeMap if they do not already exist.
     * @param x
     * @return x! as a BigInteger
     */
    public BigInteger factorial(Integer x) {
        if (_Factorial_TreeMap == null) {
            init_Factorial_TreeMap();
        }
        if (_Factorial_TreeMap.containsKey(x)) {
            return _Factorial_TreeMap.get(x);
        }
        Integer lastKey = _Factorial_TreeMap.lastKey();
        BigInteger lastFactorial = _Factorial_TreeMap.get(lastKey);
        BigInteger factorial = null;
        BigInteger value;
        Integer y;
        int key = lastKey + 1;
        for (int keys = key; keys <= x; keys++) {
            y = new Integer(keys);
            value = new BigInteger("" + keys);
            factorial = lastFactorial.multiply(value);
            _Factorial_TreeMap.put(y, factorial);
            lastFactorial = factorial;
        }
        return factorial;
    }

    /**
     * @param x
     * @param y
     * @return x^y
     */
    public static BigDecimal power(
            BigInteger x,
            int y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (y < 0) {
            return reciprocal(
                    x.pow(y),
                    decimalPlaces,
                    a_RoundingMode);
        } else {
            return new BigDecimal(x.pow(y));
        }
    }

    /**
     * @param x
     * @param y
     * @return x^y
     */
    public static BigDecimal power(
            BigInteger x,
            long y,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        if (y <= Generic_long.Integer_MAX_VALUE) {
            return power(
                    x,
                    (int) y,
                    decimalPlaces,
                    a_RoundingMode);
        }
        BigDecimal result;
        long y0 = y / Generic_long.Integer_MAX_VALUE;
        BigDecimal y0Power = power(
                x,
                y0,
                decimalPlaces,
                a_RoundingMode);
        long y1 = y - y0;
        BigDecimal y1Power = power(
                x,
                y1,
                decimalPlaces,
                a_RoundingMode);
        result = Generic_BigDecimal.multiply(
                y0Power,
                y1Power,
                decimalPlaces,
                a_RoundingMode);
        return result;
    }

    /**
     * @param x
     * @param decimalPlace
     * @param a_RoundingMode
     * @return 1/x Accurate to decimalPlace number of decimal places. If x = 0
     * then an IllegalArgumentException is thrown
     */
    public static BigDecimal reciprocal(
            BigInteger x,
            int decimalPlace,
            RoundingMode a_RoundingMode) {
        if (x.compareTo(BigInteger.ZERO) == 0) {
            throw new IllegalArgumentException(
                    "x = 0 in " + Generic_BigInteger.class
                    + ".reciprocal(BigInteger,int,RoundingMode)");
        }
        BigDecimal result = BigDecimal.ONE.divide(
                new BigDecimal(x),
                decimalPlace,
                a_RoundingMode);
        return Generic_BigDecimal.round(result, decimalPlace, a_RoundingMode);
    }

    /**
     * e^x = 1 + x/1! +x^2/2! + x^3/3! +...
     * @param x
     * @param decimalPlaces
     * @return e^x where e is like Generic_BigDecimal._E
     */
    public static BigDecimal exp(
            BigInteger x,
            BigDecimal e,
            Generic_BigDecimal a_Generic_BigDecimal,
            int decimalPlaces,
            RoundingMode a_RoundingMode) {
        BigDecimal result;
        // Deal with special cases
        if (x.compareTo(BigInteger.ZERO) == 0) {
            return BigDecimal.ONE;
        }
        if (x.compareTo(BigInteger.ONE) == 0) {
            //return getE(decimalPlaces);
            return e;
        }
        int safeDecimalPlaces;
        if (x.compareTo(BigInteger.ZERO) == -1) {
            safeDecimalPlaces = decimalPlaces + 1;
            BigDecimal exp = exp(
                    x.negate(),
                    e,
                    a_Generic_BigDecimal,
                    safeDecimalPlaces,
                    a_RoundingMode);
            result = Generic_BigDecimal.reciprocal(
                    exp,
                    decimalPlaces,
                    a_RoundingMode);
            return result;
        }
        safeDecimalPlaces = decimalPlaces + 3;
        result = BigDecimal.ONE.add(new BigDecimal(x));
        //BigDecimal e = getE(safeDecimalPlaces);
        //int maxite = 10000;
        //int maxite = decimalPlaces * 2;
        int maxite = safeDecimalPlaces;
        if (a_Generic_BigDecimal._Factorial_Generic_BigInteger == null) {
            a_Generic_BigDecimal.init_Factorial_Generic_BigInteger(maxite);
        } else {
            a_Generic_BigDecimal._Factorial_Generic_BigInteger.factorial(maxite);
        }
        BigDecimal factorial;
        BigDecimal xpow;
        BigDecimal xpowOverFactorial;
        BigDecimal tollerance = new BigDecimal(
                BigInteger.ONE, safeDecimalPlaces);
        // Use Taylor Series
        Integer factorialKey;
        BigInteger i_BigInteger;
        for (int i = 2; i < maxite; i++) {
            i_BigInteger = BigInteger.valueOf(i);
            factorialKey = new Integer(i);
            factorial = new BigDecimal(
                    a_Generic_BigDecimal._Factorial_Generic_BigInteger._Factorial_TreeMap.get(factorialKey));
            xpow = power(
                    x,
                    i,
                    safeDecimalPlaces,
                    a_RoundingMode);
            xpowOverFactorial = xpow.divide(
                    factorial, safeDecimalPlaces, a_RoundingMode);
            result = result.add(xpowOverFactorial);
            if (xpowOverFactorial.compareTo(tollerance) == -1) {
                break;
            }
        }
        return Generic_BigDecimal.round(result, decimalPlaces, a_RoundingMode);
    }
}
